@@ -1,29 +1,42 @@
 from __future__ import annotations
 
-import vapoursynth as vs
-from vsexprtools import ExprOp, combine, expect_bits, expr_func, shift_clip_multi
+from functools import partial
+
+from vsexprtools import ExprOp, combine, norm_expr
 from vskernels import Catrom
-from vsmask.edge import PrewittTCanny
+from vsmask.edge import EdgeDetect, Prewitt, PrewittTCanny
 from vsmask.util import XxpandMode, expand
 from vsrgtools import box_blur, gauss_blur, removegrain
-from vsutil import depth, get_depth, get_neutral_value, get_y, iterate, scale_value, split
+from vsrgtools.util import wmean_matrix
+from vstools import (
+    check_variable, core, depth, expect_bits, get_depth, get_neutral_value, get_peak_value, get_y, iterate,
+    scale_thresh, scale_value, shift_clip_multi, split, vs
+)
 
 __all__ = [
     'descale_detail_mask', 'descale_error_mask',
     'simple_detail_mask', 'multi_detail_mask',
-    'credit_mask'
+    'credit_mask',
+    'ringing_mask'
 ]
 
-core = vs.core
 
+def descale_detail_mask(
+    clip: vs.VideoNode, rescaled: vs.VideoNode, thr: float = 0.05,
+    inflate: int = 2, xxpand: tuple[int, int] = [4, 0]
+) -> vs.VideoNode:
+    mask = norm_expr([get_y(clip), get_y(rescaled)], 'x y - abs')
 
-def descale_detail_mask(clip: vs.VideoNode, rescaled: vs.VideoNode, threshold: float = 0.05) -> vs.VideoNode:
-    mask = expr_func([get_y(clip), get_y(rescaled)], 'x y - abs')
+    mask = mask.std.Binarize(thr * get_peak_value(mask))
 
-    mask = mask.std.Binarize(threshold)
+    if xxpand[0]:
+        mask = iterate(mask, core.std.Maximum if xxpand[0] > 0 else core.std.Minimum, xxpand[0])
 
-    mask = iterate(mask, core.std.Maximum, 4)
-    mask = iterate(mask, core.std.Inflate, 2)
+    if inflate:
+        mask = iterate(mask, core.std.Inflate, inflate)
+
+    if xxpand[1]:
+        mask = iterate(mask, core.std.Maximum if xxpand[1] > 0 else core.std.Minimum, xxpand[1])
 
     return mask.std.Limiter()
 
@@ -41,18 +54,17 @@ def descale_error_mask(
     bit_depth = get_depth(clip)
     neutral = get_neutral_value(clip)
 
-    error = expr_func([y, rescaled], 'x y - abs')
+    error = norm_expr([y, rescaled], 'x y - abs')
 
     if bwbias > 1 and chroma:
-        chroma_abs = expr_func(chroma, f'x {neutral} - abs y {neutral} - abs max')
+        chroma_abs = norm_expr(chroma, f'x {neutral} - abs y {neutral} - abs max')
         chroma_abs = Catrom().scale(chroma_abs, y.width, y.height)
 
         tv_low, tv_high = scale_value(16, 8, bit_depth), scale_value(235, 8, bit_depth)
-        bias = expr_func([y, chroma_abs], f'x {tv_high} >= x {tv_low} <= or y 0 = and {bwbias} 1 ?')
-
+        bias = norm_expr([y, chroma_abs], f'x {tv_high} >= x {tv_low} <= or y 0 = and {bwbias} 1 ?')
         bias = expand(bias, 2)
 
-        error = expr_func([error, bias], 'x y *')
+        error = norm_expr([error, bias], 'x y *')
 
     if isinstance(expands, int):
         exp1 = exp2 = exp3 = expands
@@ -115,7 +127,7 @@ def credit_mask(
         start_frame=0, thr=thr, prefilter=prefilter
     )
 
-    bits, credit_mask = expect_bits(ed_mask)
+    credit_mask, bits = expect_bits(ed_mask)
     credit_mask = iterate(credit_mask, core.std.Minimum, 6)
     credit_mask = iterate(credit_mask, lambda x: core.std.Minimum(x).std.Maximum(), 8)
     if expand:
@@ -128,7 +140,7 @@ def credit_mask(
 def simple_detail_mask(
     clip: vs.VideoNode, sigma: float | None = None, rad: int = 3, brz_a: float = 0.025, brz_b: float = 0.045
 ) -> vs.VideoNode:
-    from lvsfunc import range_mask, scale_thresh
+    from lvsfunc import range_mask
 
     brz_a = scale_thresh(brz_a, clip)
     brz_b = scale_thresh(brz_b, clip)
@@ -155,3 +167,40 @@ def multi_detail_mask(clip: vs.VideoNode, thr: float = 0.015) -> vs.VideoNode:
             iterate(general_mask, core.std.Maximum, 3).std.Maximum().std.Inflate()
         ], ExprOp.MIN), general_mask.std.Maximum()
     ], ExprOp.MIN)
+
+
+def ringing_mask(
+    clip: vs.VideoNode,
+    rad: int = 2, brz: float = 0.35,
+    thmi: float = 0.315, thma: float = 0.5,
+    thlimi: float = 0.195, thlima: float = 0.392,
+    credit_mask: vs.VideoNode | EdgeDetect = Prewitt()
+) -> vs.VideoNode:
+    assert check_variable(clip, ringing_mask)
+
+    smax = get_peak_value(clip)
+
+    thmi, thma, thlimi, thlima = (
+        scale_thresh(t, clip) for t in [thmi, thma, thlimi, thlima]
+    )
+
+    if isinstance(credit_mask, vs.VideoNode):
+        edgemask = depth(credit_mask, get_depth(clip))
+    elif isinstance(credit_mask, EdgeDetect):
+        edgemask = credit_mask.edgemask(get_y(clip))
+
+    edgemask = get_y(edgemask).std.Limiter()
+
+    light = norm_expr(edgemask, f'x {thlimi} - {thma - thmi} / {smax} *')
+    shrink = iterate(light, core.std.Maximum, rad)
+    shrink = shrink.std.Binarize(scale_thresh(brz, clip))
+    shrink = iterate(shrink, core.std.Minimum, rad)
+    shrink = iterate(shrink, partial(core.std.Convolution, matrix=wmean_matrix), 2)
+
+    strong = norm_expr(edgemask, f'x {thmi} - {thlima - thlimi} / {smax} *')
+    expand = iterate(strong, core.std.Maximum, rad)
+
+    mask = norm_expr([expand, strong, shrink], 'x y z max - 2 *')
+    mask = mask.std.Convolution(wmean_matrix)
+
+    return norm_expr(mask, 'x 2 *').std.Limiter()
